// RegKey.h: interface for the CRegKey class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_REGKEY_H__AA43F733_6EF7_11D1_9A3A_988912000000__INCLUDED_)
#define AFX_REGKEY_H__AA43F733_6EF7_11D1_9A3A_988912000000__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "auto_array.h"

/* For some stange reason, DuplicateHandle doesn't perform under Windows 95,
 * even though the docs says it should. That being the case, if REGKEY_USE_DUP
 * is defined, we use DuplicateHandle, otherwise we use ownership transferring,
 * the same as with auto_ptr.
 */
class CRegKey
{
   class Except:public exception
   {
#define REGKEY_EXCEPT_MAX_BUFFER 250
      char m_buffer[REGKEY_EXCEPT_MAX_BUFFER];
      DWORD error_number;
   public:
      Except( const char * string ) : error_number(ERROR_SUCCESS)
      {
         strncpy( m_buffer, string, REGKEY_EXCEPT_MAX_BUFFER-1 );
      }
      Except( DWORD errNum ) : error_number( errNum )
      {
         FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, NULL, errNum, 0, m_buffer,
            REGKEY_EXCEPT_MAX_BUFFER, NULL );
      }
      Except( const Except & rhs ) : error_number( rhs.error_number )
      {
         strncpy( m_buffer, rhs.m_buffer, REGKEY_EXCEPT_MAX_BUFFER );
      }
      const char * what() const { return m_buffer; }
      DWORD get_error() const { return error_number; }
#undef REGKEY_EXCEPT_MAX_BUFFER
   };

   static struct keyData {
      HKEY hKey;
      DWORD nSubKeys;
      DWORD nMaxSubkeyLen;
      DWORD nValues;
      DWORD nMaxValueName;
      DWORD nMaxValueLen;
      DWORD SecurityDesc;
   } KeyData;

   HKEY m_hKey;
#ifndef REGKEY_USE_DUP
   mutable bool m_owner;
#endif

#ifdef REGKEY_USE_DUP
   HKEY DupHandle( HKEY hKey );
#endif
   void Release() 
   {
#ifndef REGKEY_USE_DUP
      if( m_owner )
#endif
         RegCloseKey( m_hKey );

      m_hKey=NULL;
      KeyData.hKey=NULL;
   }
   void GetKeyInfo() const;
public:
	string EnumValue( DWORD index );
   string EnumSubkey( DWORD index ) const;
   CRegKey( HKEY hKey ) : m_hKey( hKey )
#ifndef REGKEY_USE_DUP
      , m_owner(true)
#endif
   {
      KeyData.hKey=NULL;
   }

   CRegKey() : m_hKey( NULL ) // Default constructor
#ifndef REGKEY_USE_DUP
      , m_owner(false)
#endif
   {
      KeyData.hKey=NULL;
   }
   
#ifdef REGKEY_USE_DUP
   CRegKey( const CRegKey & rhs ) : m_hKey( DupHandle( rhs.m_hKey ) )
   {
      KeyData.hKey=NULL;
   }
#else
   CRegKey( const CRegKey & rhs ) : m_hKey( rhs.m_hKey ), m_owner( rhs.m_owner )
   {
      KeyData.hKey=NULL;
      rhs.m_owner=false;
   }
#endif
   
   ~CRegKey()
   {
      Release();
   }

   CRegKey & operator= ( const CRegKey & rhs )
   {
      Release();
#ifdef REGKEY_USE_DUP
      m_hKey=DupHandle( rhs.m_hKey );
#else
      m_hKey=rhs.m_hKey;
      m_owner=rhs.m_owner;
      rhs.m_owner=false;
#endif

      return *this;
   }

   operator HKEY() { return m_hKey; }

   // Constructing methods
   void SetValue( const char * value, const char * data );
	void SetValue( const char * value, DWORD data );
   void SetValue( const char * value, const std::basic_string< unsigned char > & data,
      DWORD type=REG_BINARY );
   void SetValue( const char * value, const std::vector< string > & data );

	void DeleteValue( const char * value );
	void DeleteKey( const char * subKey );
	DWORD QueryValueType( const char * Value ) const;

   CRegKey Create( const char * name, REGSAM Access=KEY_ALL_ACCESS,
      DWORD dwOptions=REG_OPTION_NON_VOLATILE ) const;
   CRegKey Open( const char * name, REGSAM Access=KEY_ALL_ACCESS ) const;

   DWORD QueryNumSubkeys() const { GetKeyInfo(); return KeyData.nSubKeys; }
   DWORD QueryMaxSubkeyLen() const { GetKeyInfo(); return KeyData.nMaxSubkeyLen; }
   DWORD QueryNumValues() const { GetKeyInfo(); return KeyData.nValues;}
   DWORD QueryMaxValueName() const { GetKeyInfo(); return KeyData.nMaxValueName;}
   DWORD QueryMaxValueLen() const { GetKeyInfo(); return KeyData.nMaxValueLen;}
   DWORD QuerySecurityDesc() const { GetKeyInfo(); return KeyData.SecurityDesc;}

   DWORD QueryValue( const char * name, DWORD Default ) const;
   string QueryValue( const char * name, const string &Default ) const;
   std::basic_string< unsigned char > QueryValue( const char * name ) const;
   std::vector< string > QueryValue( const char * name, const std::vector< string > &Default )
      const; // Used for MultiSZ query
};

extern const CRegKey HKCR; // Classes Root
extern const CRegKey HKCU; // Current User
extern const CRegKey HKLM; // Local Machine
extern const CRegKey HKU;  // Users
extern const CRegKey HKPD; // Performance Data
extern const CRegKey HKCC; // Current Config
extern const CRegKey HKDD; // Dynamic Data

#endif // !defined(AFX_REGKEY_H__AA43F733_6EF7_11D1_9A3A_988912000000__INCLUDED_)
