// ui.cpp : Displays the actual eyes.
//

// This file is part of the Xeyes for windows program (c) copyright 1998 Shachar Shemesh
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation version 2 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//                                             
#include "stdafx.h"
#include "ui.h"
#include "resource.h"
#include "mouse/mouse.h"

static const int EYE_DIAMETER=2;

// Store the eyes center.
static int xLeftCenter, xRightCenter, yCenter;
// The eyes radius
static int xEyeSize, yEyeSize;
// Inner radius - eye balls are only painted up to this radius
static int xInnerSize, yInnerSize;
static int eyeBallRadius;
// The Regions for the left and right eye
HRGN hrgnRight=CreateRectRgn(0,0,0,0), hrgnLeft=CreateRectRgn(0,0,0,0);
HBRUSH hbrBlack, hbrWhite;
HPEN hpenBlack, hpenWhite;
bool g_bTransparent;
static HDC hdc;
static HPEN hOriginalPen;
static HMENU hMenu;

static void PaintEyeBalls( HDC hdc, int xMouse, int yMouse );
static void SetTopmost( HWND hwnd, HMENU hMenu, bool bTop );
static void SetTransparent( HWND hwnd, HMENU hMenu, bool bTransparent );
static void CalcWindowRgn( HWND hwnd );
static void ReleaseResources( HWND hwnd );

LRESULT CALLBACK WndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT ret_val=ERROR_SUCCESS;
   static LRESULT real_hit_test;

	switch( uMsg )
	{
   case  WM_CREATE:
      hbrBlack=CreateSolidBrush(BLACK);
      hbrWhite=CreateSolidBrush(WHITE);
      hpenWhite=CreatePen( PS_SOLID, 0, WHITE );
      hpenBlack=CreatePen( PS_SOLID, 0, BLACK );
      InstallHook( hwnd );
      hdc=GetDC(hwnd);
      hOriginalPen=reinterpret_cast<HPEN>(SelectObject( hdc, hpenWhite ));
      hMenu=LoadMenu( NULL, MAKEINTRESOURCE( IDR_POPMENU ) );
      {
         CRegKey hKey=HKCU.Create(XEYES_REGISTRY_KEY);

         SetTopmost( hwnd, hMenu, hKey.QueryValue("Topmost", false )!=false );
         SetTransparent( hwnd, hMenu, hKey.QueryValue("Transparent", false )!=false );
      }
      break;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc=BeginPaint( hwnd, &ps );
         HANDLE hOldBrush=SelectObject( hdc, hbrBlack );

         // Outsides
         // Right eye (the one on the left)
         Ellipse( hdc, xLeftCenter-xEyeSize, yCenter-yEyeSize, xLeftCenter+xEyeSize, yCenter+yEyeSize );
         // Left eye (the one on the right)
         Ellipse( hdc, xRightCenter-xEyeSize, yCenter-yEyeSize, xRightCenter+xEyeSize, yCenter+yEyeSize );

         // Insides
         SelectObject( hdc, hOldBrush );
         // Right eye (the one on the left)
         Ellipse( hdc, xLeftCenter-xEyeSize+EYE_DIAMETER, yCenter-yEyeSize+EYE_DIAMETER,
            xLeftCenter+xEyeSize-EYE_DIAMETER, yCenter+yEyeSize-EYE_DIAMETER );
         // Left eye (the one on the right)
         Ellipse( hdc, xRightCenter-xEyeSize+EYE_DIAMETER, yCenter-yEyeSize+EYE_DIAMETER,
            xRightCenter+xEyeSize-EYE_DIAMETER, yCenter+yEyeSize-EYE_DIAMETER );

			EndPaint( hwnd, &ps );
		}

      PaintEyeBalls( hdc, (xLeftCenter+xRightCenter)/2, yCenter );
		break;
   case WM_SIZE:
      // Split the area to two 45% areas, and a 10% gap in the middle.
      xEyeSize=495*LOWORD(lParam)/2000;
      xInnerSize=30*LOWORD(lParam)/200;
      // Vertically we can saftly take the entire area
      yEyeSize=HIWORD(lParam)/2;
      yInnerSize=60*HIWORD(lParam)/200;

      xLeftCenter=xEyeSize;
      xRightCenter=LOWORD(lParam)-xEyeSize;
      yCenter=HIWORD(lParam)/2;

      eyeBallRadius=min(xEyeSize,yEyeSize)/4;

      DeleteObject( hrgnLeft );
      DeleteObject( hrgnRight );

      hrgnLeft=CreateEllipticRgn( xLeftCenter-xInnerSize, yCenter-yInnerSize, xLeftCenter+xInnerSize,
         yCenter+yInnerSize );
      hrgnRight=CreateEllipticRgn( xRightCenter-xInnerSize, yCenter-yInnerSize, xRightCenter+xInnerSize,
         yCenter+yInnerSize );

      if( g_bTransparent )
      {
         CalcWindowRgn( hwnd );
      }

      break;
   case WMU_MOUSEMOVE:
      {
         POINT pt;
         
         pt.x=LOWORD( lParam );
         pt.y=HIWORD( lParam );
         ScreenToClient( hwnd, &pt );
         PaintEyeBalls( hdc, pt.x, pt.y );
      }
      break;
   case WM_ENDSESSION:
      if( wParam )
         ReleaseResources( hwnd );
      break;
   case WM_DESTROY:
      ReleaseResources( hwnd );

      PostQuitMessage(0);
      break;
   case WM_COMMAND:
      {
         switch( LOWORD(wParam) )
         {
         case ID_ONTOP:
            {
               MENUITEMINFO mii;
               mii.cbSize=sizeof( mii );
               mii.fMask=MIIM_STATE;
               GetMenuItemInfo( hMenu, ID_ONTOP, FALSE, &mii );

               SetTopmost( hwnd, hMenu, (mii.fState&MFS_CHECKED)==0 );
            }
            break;
         case ID_TRANSPARENT_WINDOW:
            {
               SetTransparent( hwnd, hMenu, !g_bTransparent );
            }
            break;
         case IDCLOSE:
            DestroyWindow(hwnd);
            break;
         }
      }
      break;
   case WM_ENTERMENULOOP:
      EnableHook( FALSE );
      break;
   case WM_EXITMENULOOP:
      EnableHook( TRUE );
      break;
   case WM_RBUTTONDOWN:
   case WM_NCRBUTTONDOWN:
      {

         if( real_hit_test==HTCLIENT )
         {
            // We want to open the menu where the mouse is, but the coordinates for TrackPopup are 
            // screen coordinates. We need to translate.
            POINT pt={ LOWORD(lParam), HIWORD(lParam) };
            // The following line is a relic from the non-NC days

            if( !g_bTransparent )
               ClientToScreen( hwnd, &pt );
            
            // Let's open a submenu of options
            TrackPopupMenu( GetSubMenu( hMenu, 0 ), TPM_LEFTALIGN|TPM_TOPALIGN|TPM_RIGHTBUTTON,
               pt.x, pt.y, 0, hwnd, NULL );
         } else
         {
            ret_val=DefWindowProc( hwnd, uMsg, wParam, lParam );
         }
      }
      break;
   case WM_NCHITTEST:
      {
         real_hit_test=ret_val=DefWindowProc( hwnd, uMsg, wParam, lParam );

         if( g_bTransparent && ret_val==HTCLIENT )
            ret_val=HTCAPTION;
      }
      break;
	default:
		ret_val=DefWindowProc( hwnd, uMsg, wParam, lParam );
		break;
	}

	return ret_val;
}

static void PaintEyeBalls( HDC hdc, int xMouse, int yMouse )
{
   static lastXmouseRight=0, lastYmouseRight=0;
   static lastXmouseLeft=0, lastYmouseLeft=0;

   // Clip to only use internal left eye
   HANDLE hOldBrush=SelectObject( hdc, hbrWhite );

   Ellipse( hdc, lastXmouseLeft-eyeBallRadius, lastYmouseLeft-eyeBallRadius, lastXmouseLeft+eyeBallRadius,
      lastYmouseLeft+eyeBallRadius );
   if( PtInRegion( hrgnLeft, xMouse, yMouse ) )
   {
      lastXmouseLeft=xMouse;
      lastYmouseLeft=yMouse;
   } else
   {
      double x=xMouse-xLeftCenter;
      double y=yMouse-yCenter;
      double D=x*x/(xInnerSize*xInnerSize)+y*y/(yInnerSize*yInnerSize);

      lastXmouseLeft=x/sqrt(D)+xLeftCenter;
      lastYmouseLeft=y/sqrt(D)+yCenter;
   }

   Ellipse( hdc, lastXmouseRight-eyeBallRadius, lastYmouseRight-eyeBallRadius, lastXmouseRight+eyeBallRadius,
      lastYmouseRight+eyeBallRadius );

   if( PtInRegion( hrgnRight, xMouse, yMouse ) )
   {
      lastXmouseRight=xMouse;
      lastYmouseRight=yMouse;
   } else
   {
      double x=xMouse-xRightCenter;
      double y=yMouse-yCenter;
      double D=x*x/(xInnerSize*xInnerSize)+y*y/(yInnerSize*yInnerSize);

      lastXmouseRight=x/sqrt(D)+xRightCenter;
      lastYmouseRight=y/sqrt(D)+yCenter;
   }

   SelectObject( hdc, hbrBlack );

   Ellipse( hdc, lastXmouseLeft-eyeBallRadius, lastYmouseLeft-eyeBallRadius, lastXmouseLeft+eyeBallRadius,
      lastYmouseLeft+eyeBallRadius );
   
   Ellipse( hdc, lastXmouseRight-eyeBallRadius, lastYmouseRight-eyeBallRadius, lastXmouseRight+eyeBallRadius,
      lastYmouseRight+eyeBallRadius );

   SelectObject( hdc, hOldBrush );
}

static void SetTopmost( HWND hwnd, HMENU hMenu, bool bTop )
{
   MENUITEMINFO mii;
   mii.cbSize=sizeof( mii );
   mii.fMask=MIIM_STATE;
   mii.fState=bTop?MFS_CHECKED:MFS_UNCHECKED;
   SetMenuItemInfo( hMenu, ID_ONTOP, FALSE, &mii );
   
   SetWindowPos( hwnd, bTop?HWND_TOPMOST:HWND_NOTOPMOST, 0, 0, 0, 0,
      SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE|SWP_NOOWNERZORDER );

   HKCU.Create(XEYES_REGISTRY_KEY).SetValue( "Topmost", bTop );
}

static void SetTransparent( HWND hwnd, HMENU hMenu, bool bTransparent )
{
   MENUITEMINFO mii;
   mii.cbSize=sizeof( mii );
   mii.fMask=MIIM_STATE;
   mii.fState=bTransparent?MFS_CHECKED:MFS_UNCHECKED;
   SetMenuItemInfo( hMenu, ID_TRANSPARENT_WINDOW, FALSE, &mii );

   g_bTransparent=bTransparent;

   if( bTransparent )
      CalcWindowRgn( hwnd );
   else
      SetWindowRgn( hwnd, NULL, TRUE );

   HKCU.Create(XEYES_REGISTRY_KEY).SetValue( "Transparent", bTransparent );
}

static void CalcWindowRgn( HWND hwnd )
{
   RECT wndRect;
   GetWindowRect( hwnd, &wndRect );
   
   POINT pt;
   pt.x=0;
   pt.y=0;
   ClientToScreen( hwnd, &pt );
   
   int xoff=pt.x-wndRect.left;
   int yoff=pt.y-wndRect.top;
   
   HRGN hWinRgn=CreateEllipticRgn( xoff+xLeftCenter-xEyeSize, yoff+yCenter-yEyeSize,
      xoff+xLeftCenter+xEyeSize+1, yoff+yCenter+yEyeSize+1 );
   
   HRGN hOtherRgn=CreateEllipticRgn( xoff+xRightCenter-xEyeSize, yoff+yCenter-yEyeSize,
      xoff+xRightCenter+xEyeSize+1, yoff+yCenter+yEyeSize+1 );
   
   CombineRgn( hWinRgn, hWinRgn, hOtherRgn, RGN_OR );
   DeleteObject( hOtherRgn );
   
   SetWindowRgn( hwnd, hWinRgn, TRUE );
}

static void ReleaseResources( HWND hwnd )
{
   try
   {
      RECT winRect;
      GetWindowRect( hwnd, &winRect );
      CRegKey xeyesKey=HKCU.Create(XEYES_REGISTRY_KEY);
      xeyesKey.SetValue("Left", winRect.left );
      xeyesKey.SetValue("Top", winRect.top );
      xeyesKey.SetValue("Width", winRect.right-winRect.left );
      xeyesKey.SetValue("Height", winRect.bottom-winRect.top );
   } catch(...)
   {
   }
   
   SelectObject( hdc, hOriginalPen );
   ReleaseDC( hwnd, hdc );
   
   DeleteObject( hrgnLeft );
   DeleteObject( hrgnRight );
   DeleteObject( hbrWhite );
   DeleteObject( hbrBlack );
   DeleteObject( hpenWhite );
   DeleteObject( hpenBlack );
   RemoveHook();
   
   DestroyMenu( hMenu );
}
